#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import collections
import warnings
from logging import Logger as BaseLogger, NOTSET
from types import MappingProxyType

from .._logger import Logger

# fmt: off
level_to_severity = MappingProxyType({
     0:   0,
    10: 200,
    20: 400,
    30: 600,
    40: 800,
    50: 900,
})
# fmt: on


class LegacyLogger(BaseLogger):
    """Drop-in replacement of ``logging.Logger`` to support handling messages from existing libraries.

    Provides the same interface as ``logging.Logger`` but behaves like :class:`hakisto.Logger`.

    Use ``logging.setLoggerClass()`` to enable.

    .. code:: python

       from logging import setLoggerClass
       from hakisto.legacy import LegacyLogger

       setLoggerClass(LegacyLogger)

    or

    .. code:: python

       from hakisto.legacy import enable_legacy_logger

       enable_legacy_logger()
    """

    def __init__(self, name: str, level: int = NOTSET):
        super(LegacyLogger, self).__init__(name, level)
        self._logger = Logger(name)

    def setLevel(self, level):
        """
        Set the logging level of this logger.  level must be an int or a str.
        """
        super(LegacyLogger, self).setLevel(level)
        self._logger.set_severity(level_to_severity.get(self.level, 0))

    @staticmethod
    def _convert_legacy_message(msg, args) -> str:
        if all((args,
                len(args) == 1,
                isinstance(args[0], collections.abc.Mapping),
                args[0])):
            args = args[0]
        args = args
        if args:
            msg = str(msg) % args
        return msg

    def debug(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'DEBUG'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        self._logger.debug(self._convert_legacy_message(msg, args), **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.info("Houston, we have a %s", "notable problem", exc_info=1)
        """
        self._logger.info(self._convert_legacy_message(msg, args), **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.warning("Houston, we have a %s", "bit of a problem", exc_info=1)
        """
        self._logger.warning(self._convert_legacy_message(msg, args), **kwargs)

    def warn(self, msg, *args, **kwargs):
        warnings.warn("The 'warn' method is deprecated, use 'warning' instead", DeprecationWarning, 2)
        self._logger.warning(self._convert_legacy_message(msg, args), **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'ERROR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.error("Houston, we have a %s", "major problem", exc_info=1)
        """
        self._logger.error(self._convert_legacy_message(msg, args), **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        """
        Convenience method for logging an ERROR with exception information.
        """
        self._logger.critical(self._convert_legacy_message(msg, args), **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'CRITICAL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.critical("Houston, we have a %s", "major disaster", exc_info=1)
        """
        self._logger.critical(self._convert_legacy_message(msg, args), **kwargs)

    def fatal(self, msg, *args, **kwargs):
        """
        Don't use this method, use critical() instead.
        """
        self._logger.critical(self._convert_legacy_message(msg, args), **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """
        Log 'msg % args' with the integer severity 'level'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.log(level, "We have a %s", "mysterious problem", exc_info=1)
        """
        self._logger.log(level_to_severity.get(level, 0),
                         self._convert_legacy_message(msg, args),
                         **kwargs)



import inspect

from hakisto import logger

__all__ = ['hakisto_file', 'hakisto_process_file', 'hakisto_process_severity', 'hakisto_severity']


logger._Logger.register_excluded_source_file(inspect.currentframe().f_code.co_filename)

try:
    import click
except ImportError as e:
    logger.critical("package 'click' not found")
    raise e from None
else:
    from ._click import hakisto_file, hakisto_process_file, hakisto_process_severity, hakisto_severity



import os
import sys
import typing as t

import click
from click.decorators import _param_memo

from hakisto import logger
from .severity import severity_names


if t.TYPE_CHECKING:
    import typing_extensions as te

    P = te.ParamSpec("P")

R = t.TypeVar("R")
T = t.TypeVar("T")
_AnyCallable = t.Callable[..., t.Any]
FC = t.TypeVar("FC", bound=t.Union[_AnyCallable, click.Command])


def hakisto_severity(
        *param_decls: str,
        default: str = None,
        choices: t.Sequence[str] = None,
        case_sensitive: bool = False,
        show_default: bool = True,
        cls: t.Optional[t.Type[click.Option]] = None,
        **attrs: t.Any) -> t.Callable[[FC], FC]:
    """Attaches HAKISTO SEVERITY option to the command.

    This is always a ``click.Choice`` option, with the default taken from HAKISTO_SEVERITY or ``INFO``.

    The default options are the Hakisto Severities (not case-sensitive).

    All positional arguments are  passed as parameter declarations to :class:`Option`,
    if none are provided, defaults to ``--log, --log-severity``.

    Keyword arguments not described below are passed as keyword arguments to :func:`click.Option`,
    arguments are forwarded unchanged (except ``cls``).

    :param default: default Hakisto Severity.
    :param choices: List of permitted Hakisto Severities.
    :param case_sensitive:
    :param show_default:
    :param cls: the option class to instantiate.  This defaults to :class:`Option`.
    :param param_decls: Passed as positional arguments to the constructor of ``cls``.
    :param attrs: Passed as keyword arguments to the constructor of ``cls``.
    """
    if cls is None:
        cls = click.Option
    if not param_decls:
        param_decls = ('--log', '--log-severity', 'log_severity')

    if not default:
        default = os.getenv("HAKISTO_SEVERITY", "INFO")
    attrs['default'] = default.upper()
    attrs['show_default'] = show_default
    if not choices:
        choices = severity_names.keys()
    choices = [i.upper() for i in choices]
    attrs['type'] = click.Choice(choices, case_sensitive=case_sensitive)
    if 'help' not in attrs:
        attrs['help'] = "Minimum Logging Severity"

    def decorator(f: FC) -> FC:
        _param_memo(f, cls(param_decls, **attrs))
        return f

    return decorator


def hakisto_file(
        *param_decls: str,
        default: bool = True,
        show_default: bool = True,
        cls: t.Optional[t.Type[click]] = None,
        **attrs: t.Any) -> t.Callable[[FC], FC]:
    """Attaches option enable or disable logging to a file to the command.

    All positional arguments are  passed as parameter declarations to :class:`Option`,
    if none are provided, defaults to ``--log-file/--no-log-file``.

    Keyword arguments not described below are passed as keyword arguments to :func:`click.Option`,
    arguments are forwarded unchanged (except ``cls``).

    :param default:
    :param show_default:
    :param cls: the option class to instantiate.  This defaults to :class:`Option`.
    :param param_decls: Passed as positional arguments to the constructor of ``cls``.
    :param attrs: Passed as keyword arguments to the constructor of ``cls``.
    """
    if cls is None:
        cls = click.Option

    if not param_decls:
        param_decls = ('--log-file/--no-log-file',)
    attrs['default'] = default
    attrs['show_default'] = show_default
    if 'help' not in attrs:
        attrs['help'] = "Log to file?"

    def decorator(f: FC) -> FC:
        _param_memo(f, cls(param_decls, **attrs))
        return f

    return decorator


def hakisto_process_severity(log_severity: str) -> None:
    """Process the Hakisto Severity option.

    :param log_severity:
    :type log_severity: str
    """
    logger.set_severity(severity_names.get(log_severity, 0))


def hakisto_process_file(log_file: bool) -> None:
    """Process the Hakisto File option.

    :param log_file:
    :type log_file: bool
    """
    if not log_file:
        logger.file_handler.severity = sys.maxsize
